from dataclasses import dataclass
from typing import Any, ClassVar, Dict, List

from ..exc import NotEnoughAddresses
from ..types import CoverageZone, OrderAddress, Package
from .base import Resource


@dataclass
class Budget(Resource):
    """
    Based on: http://docs.ivoy.mx/express/#cotización
    """

    _endpoint: ClassVar[str] = '/api/order/validZipCode/json/web'
    _response: Dict[str, Any]
    distance: float
    price: float
    eta: float
    addresses: List[OrderAddress]
    zone: CoverageZone
    package_type: Package

    @classmethod
    def create(cls, addresses: List[OrderAddress], **metadata) -> 'Budget':
        if len(addresses) < 2:
            raise NotEnoughAddresses(-111, "2 addresses minimum per budget.")
        address_array = [address.to_dict() for address in addresses]
        body = dict(data=dict(bOrder=dict(orderAddresses=address_array)))
        resp = cls._client.post(cls._endpoint, json=body)
        data = resp['data']
        return cls(
            _response=resp,
            distance=data['distance'],
            price=data['price'],
            eta=data['eta'],
            zone=CoverageZone(data['zone']['idZone']),
            package_type=Package(data['packageType']['idPackageType']),
            addresses=addresses,
        )



import json
from dataclasses import dataclass
from enum import Enum, unique
from typing import Optional


@unique
class CoverageZone(Enum):
    CDMX = 1
    GDL = 2


@unique
class Vehicle(Enum):
    bike_whitout_box = 1
    bike_with_box = 2
    moto_without_box = 3
    moto_box_medium = 5
    moto_box_big = 6


@unique
class Package(Enum):
    envelope = 1
    box_medium = 2
    box_big = 3


@unique
class PaymentMethod(Enum):
    credit_prepaid = 4


@unique
class OrderStatus(Enum):
    created = 2
    round_robin = 3
    available = 4
    in_process = 5
    finalized = 6
    cancelled_by_customer = 7
    cancelled_by_user = 8
    devolutino = 9
    finalized_by_absence = 10
    finalized_with_shelter = 11
    cancelled_by_user_with_collection = 12


@dataclass
class OrderAddress:
    is_pickup: int
    is_source: int
    phone: Optional[str]
    external_number: Optional[str]
    latitude: str
    longitude: str
    neighborhood: str
    street: str
    zip_code: str
    comment: str = 'N/A'
    internal_number: Optional[str] = None
    person_approved: Optional[str] = None
    id_address: Optional[int] = None

    def to_dict(self) -> dict:
        return dict(
            isPickup=self.is_pickup,
            isSource=self.is_source,
            comment=self.comment,
            personApproved=self.person_approved,
            phone=self.phone,
            address=dict(
                idAddress=self.id_address,
                externalNumber=self.external_number,
                internalNumber=self.internal_number,
                latitude=self.latitude,
                longitude=self.longitude,
                neighborhood=self.neighborhood,
                street=self.street,
                zipCode=self.zip_code,
            ),
        )

    def to_json(self) -> str:
        return json.dumps(self.to_dict())



from dataclasses import dataclass
from typing import ClassVar

from .base import Resource


@dataclass
class OrderSharing(Resource):
    """
    This is based in iVoy App
    """

    _endpoint: ClassVar[str] = '/api/orderSharing/createSharingURL/json/web'

    id: str
    tracking_url: str

    @classmethod
    def get_tracking_url(cls, identity_id: str) -> 'OrderSharing':
        json_data = dict(data=dict(bOrder=dict(idOrder=identity_id)))
        resp = cls._client.post(cls._endpoint, json=json_data)
        return cls(id=identity_id, tracking_url=resp['data']['trackingURL'])

    @staticmethod
    def _json_data(identity_id: str) -> dict:
        json_data = dict(data=dict(bOrder=dict(idOrder=identity_id)))
        return json_data



import os
from typing import Any, ClassVar, Dict, Optional

from requests import Response, Session

from .exc import ExpiredTokens, IvoyException
from .resources import Budget, Order, OrderSharing, Resource

API_URL = os.environ['IVOY_URL']
WEB_URL = os.environ['IVOY_WEB_URL']


class Client:

    base_url: ClassVar[str] = API_URL
    web_url: ClassVar[str] = WEB_URL
    auth_user: str
    auth_password: str
    web_auth_user: str
    web_auth_password: str
    ivoy_user: str
    ivoy_password: str
    token: Optional[str]
    web_token: Optional[str]
    session: Session

    # resources
    budget: ClassVar = Budget
    order: ClassVar = Order
    order_sharing: ClassVar = OrderSharing

    def __init__(
        self,
        auth_user: Optional[str] = None,
        auth_password: Optional[str] = None,
        ivoy_user: Optional[str] = None,
        ivoy_password: Optional[str] = None,
        web_auth_user: Optional[str] = None,
        web_auth_password: Optional[str] = None,
    ):
        self.session = Session()
        self.auth_user = auth_user or os.environ['AUTH_USER']
        self.auth_password = auth_password or os.environ['AUTH_PASS']
        self.web_auth_user = web_auth_user or os.environ['WEB_AUTH_USER']
        self.web_auth_password = (
            web_auth_password or os.environ['WEB_AUTH_PASS']
        )
        self.ivoy_user = ivoy_user or os.environ['IVOY_USER']
        self.ivoy_password = ivoy_password or os.environ['IVOY_PASS']
        self.token = self.get_token()
        self.web_token = self.get_token(True)
        Resource._client = self

    def get_token(self, web_token: bool = False) -> str:
        url = f'{self.base_url}/api/login/loginClient/json/web'
        if web_token:
            auth = (self.web_auth_user, self.web_auth_password)
        else:
            auth = (self.auth_user, self.auth_password)
        response = self.session.request(
            'POST',
            url,
            auth=auth,
            json={
                'data': {
                    'systemRequest': {
                        'user': self.ivoy_user,
                        'password': self.ivoy_password,
                    }
                }
            },
        )
        self._check_response(response)
        data = response.json()
        return data['token']['access_token']

    def post(self, endpoint: str, **kwargs: Any) -> Dict[str, Any]:
        return self.request('post', endpoint, **kwargs)

    def put(self, endpoint: str, **kwargs: Any) -> Dict[str, Any]:
        return self.request('put', endpoint, **kwargs)

    def request(
        self, method: str, endpoint: str, **kwargs: Any,
    ) -> Dict[str, Any]:
        if 'orderSharing' in endpoint:
            url = self.web_url + endpoint
        else:
            url = self.base_url + endpoint
        headers = self.create_headers(endpoint)
        response = self.session.request(method, url, headers=headers, **kwargs)

        try:
            self._check_response(response)
        except ExpiredTokens:
            # Try to generate Tokens again and retry request
            self.token = self.get_token()
            self.web_token = self.get_token(True)
            headers = self.create_headers(endpoint)
            response = self.session.request(
                method, url, headers=headers, **kwargs
            )
            self._check_response(response)
        except IvoyException as ivoyexc:
            raise IvoyException(ivoyexc.code, ivoyexc.message)
        except Exception as e:
            raise Exception(e)

        return response.json()

    def create_headers(self, endpoint: str) -> Dict[str, Any]:
        if 'orderSharing' in endpoint:
            return {
                'Content-type': 'application/json',
                'Token': self.web_token,
            }
        else:
            return {
                'Content-type': 'application/json',
                'Token': self.token,
            }

    @staticmethod
    def _check_response(response: Response) -> None:
        if response.ok:
            json = response.json()
            if 'code' in json and json['code'] == 0:
                return
            elif 'code' in json and json['code'] == -156:
                # Tokens expired. Based on http://docs.ivoy.mx/express/#errores
                # All error codes come from API as negative numbers.
                raise ExpiredTokens(156, 'Tokens have expired.')
            else:
                raise IvoyException(
                    json['code'], 'iVoy API error: {}'.format(json['message'])
                )
        response.raise_for_status()



from importlib.machinery import SourceFileLoader

from setuptools import find_packages, setup

version = SourceFileLoader('version', 'ivoy/version.py').load_module()

test_requires = [
    'pytest',
    'pytest-vcr',
    'pycodestyle',
    'pytest-cov',
    'black',
    'isort[pipfile]',
    'flake8',
    'mypy',
    'iso8601',
]

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='ivoy',
    version=version.__version__,
    author='Cuenca',
    author_email='dev@cuenca.com',
    description='Client library for iVoy',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/cuenca-mx/ivoy-python',
    packages=find_packages(),
    include_package_data=True,
    package_data=dict(ivoy=['py.typed']),
    python_requires='>=3.6',
    install_requires=[
        'dataclasses>=0.6;python_version<"3.7"',
        'requests>=2.22.0,<3.0.0',
        'iso8601>=0.1.12,<0.2.0',
    ],
    setup_requires=['pytest-runner'],
    tests_require=test_requires,
    extras_require=dict(test=test_requires),
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)



class IvoyException(Exception):
    """Generic iVoy API exception"""

    def __init__(self, code, message):
        self.code = code
        self.message = message

    def __str__(self):
        return f'Code: {self.code} - Message: {self.message}'


class ExpiredTokens(IvoyException):
    """API Tokens Expired"""


class NotEnoughAddresses(IvoyException):
    """Addresses Needed to create Budget/Order"""



import datetime as dt
from dataclasses import dataclass
from typing import Any, ClassVar, Dict, List, Optional

from ..exc import NotEnoughAddresses
from .base import Resource

from ..types import (  # noqa isort:skip
    CoverageZone,
    OrderAddress,
    OrderStatus,
    Package,
    PaymentMethod,
    Vehicle,
)


@dataclass
class Order(Resource):
    """
    Based on: http://docs.ivoy.mx/express/#órdenes
    """

    _endpoint: ClassVar[str] = '/api/order'
    _response: Dict[str, Any]
    id: Optional[int]
    date_add: dt.datetime
    date_edit: dt.datetime
    discount: float
    price: float
    eta: float
    total: float
    addresses: List[OrderAddress]
    payment_method: PaymentMethod
    zone: Optional[CoverageZone]
    order_status: OrderStatus
    order_type: Optional[int]
    package_type: Optional[Package]
    payment_status: Optional[int]
    vehicle: Optional[Vehicle]

    @classmethod
    def create(
        cls,
        addresses: List[OrderAddress],
        package: Package,
        payment: PaymentMethod,
    ) -> 'Order':
        if len(addresses) < 2:
            raise NotEnoughAddresses(-111, "2 addresses minimum per order.")
        endpoint = f'{cls._endpoint}/newOrder/json/web'
        address_array = [address.to_dict() for address in addresses]
        json_data = cls._create_json(address_array, package, payment)
        resp = cls._client.put(endpoint, json=json_data)
        return cls._to_object(resp, addresses)

    @classmethod
    def retrieve(cls, identity_id: str) -> 'Order':
        endpoint = f'{cls._endpoint}/getById/json/web'
        json_data = cls._update_json(identity_id)
        resp = cls._client.post(endpoint, json=json_data)
        return cls._to_object(resp)

    @classmethod
    def cancel(cls, identity_id: str) -> 'Order':
        endpoint = f'{cls._endpoint}/cancelOrder/json/web'
        json_data = cls._update_json(identity_id)
        resp = cls._client.put(endpoint, json=json_data)
        return cls._to_object(resp)

    @staticmethod
    def _to_object(
        response: dict, addresses: List[OrderAddress] = [],
    ) -> 'Order':
        data = response['data']
        if addresses == []:
            for order_addr in data['orderAddresses']:
                addresses.append(
                    OrderAddress(
                        is_pickup=order_addr['isPickup'],
                        is_source=order_addr['isSource'],
                        phone=order_addr['phone'],
                        external_number=order_addr['address'][
                            'externalNumber'
                        ],
                        latitude=order_addr['address']['latitude'],
                        longitude=order_addr['address']['longitude'],
                        neighborhood=order_addr['address']['neighborhood'],
                        street=order_addr['address']['street'],
                        zip_code=order_addr['address']['zipCode'],
                        person_approved=order_addr['personApproved'],
                        id_address=order_addr['address']['idAddress'],
                    )
                )
        return Order(
            _response=response,
            id=data['idOrder'],
            date_add=dt.datetime.fromtimestamp(data['dateAdd'] // 1000),
            date_edit=dt.datetime.fromtimestamp(data['dateEdit'] // 1000),
            discount=data['discount'],
            price=data['price'],
            eta=data['eta'],
            total=data['total'],
            addresses=addresses,
            payment_method=PaymentMethod(
                data['paymentMethod']['idPaymentMethod']
            ),
            zone=CoverageZone(data['zone']['idZone'])
            if 'zone' in data
            else None,
            order_status=OrderStatus(data['orderStatus']['idOrderStatus']),
            order_type=data['orderType']['idOrderType'],
            package_type=Package(data['packageType']['idPackageType'])
            if 'packageType' in data
            else None,
            payment_status=data['paymentStatus']['idPaymentStatus']
            if 'paymentStatus' in data
            else None,
            vehicle=Vehicle(data['vehicle']['idVehicle'])
            if 'vehicle' in data
            else None,
        )

    @staticmethod
    def _create_json(
        address_array: list, package: Package, payment: PaymentMethod,
    ) -> dict:
        json_data = dict(
            data=dict(
                bOrder=dict(
                    device='api',
                    orderType=dict(idOrderType=1),
                    packageType=dict(idPackageType=package.value),
                    paymentMethod=dict(idPaymentMethod=payment.value),
                    orderAddresses=address_array,
                )
            )
        )
        return json_data

    @staticmethod
    def _update_json(identity_id: str) -> dict:
        json_data = dict(data=dict(bOrder=dict(idOrder=identity_id)))
        return json_data
